%module AggTcl

// Note: this module's external interface will be inserted in a namespace called Agg::


//  swig-wrapper for AggKit library.
//  Note this wrapper is specialized for interfacing Tcl only.

%{
	#include "AggKit.h"
	#include <tk.h>  // just for names of colors

	typedef agg::srgba8       Color;  // srgba8 is rgba8T<sRGB>
	typedef agg::rect_d       RectD;  // rect_d is Rect<double> 
	typedef agg::trans_affine Transformations;
	
	class AggTcl : public AggKit {
		private:
			int m_updateScheduled;
			Tcl_Interp *m_savedInterp;
			char *m_savedUpdateScript;
			void scheduleUpdate() {
				if ( m_savedUpdateScript && !m_updateScheduled ) {
					m_updateScheduled = 1;
// to do, use Tcl_EvalObj .. faster ...
					Tcl_Eval(m_savedInterp, m_savedUpdateScript);
				}
			};
		public:
			AggTcl() { 
				m_updateScheduled = 0;
				m_savedInterp = NULL;
				m_savedUpdateScript = NULL;
			};
			~AggTcl() { if( m_savedUpdateScript ) free(m_savedUpdateScript); };
			void updated() { m_updateScheduled = 0; };
			 // extend ..
			void render(bool fillColor) {
				 // NOTE:: AggKit::render() must be virtual for late binding !
				AggKit::render(fillColor);
				scheduleUpdate();
			};
			void clear (Color c) {
				AggKit::clear(c);
				scheduleUpdate();			
			};
			 // note: this method should not be exported to swig
			void renderImage(const AggPixmap& img, int x1, int y1, int x2, int y2, const double* parl) {
				AggKit::renderImage(img, x1, y1, x2, y2, parl);                        
				scheduleUpdate();			
			}

			void blendImage(AggPixmap& img, int imgX1, int imgY1, int imgX2, int imgY2, double dstX, double dstY, unsigned alpha) {
				AggKit::blendImage(img,imgX1,imgY1,imgX2,imgY2,dstX,dstY,alpha);
				scheduleUpdate();			
			}
			void blendImage(AggPixmap& img, double dstX, double dstY, unsigned alpha) {
				AggKit::blendImage(img,dstX,dstY,alpha);
				scheduleUpdate();			
			}
							
			void setUpdateScript(Tcl_Interp * interp, const char *script) {
				m_savedInterp = interp;
				if( m_savedUpdateScript ) free(m_savedUpdateScript);
				m_savedUpdateScript = _strdup(script);
			};


	};
%}

/* ========================================================================= */
		%typemap(in,numinputs=0) Tcl_Interp * interp {
  			$1 = interp;
		}

		// This tells SWIG to treat in-param "Color" as a special case.
		// i.e.  We have C methods with an expected "Color" parameter,
		//   and in Tcl we should pass just a string like the following
		//       "lightblue"    (a symbolic name)
		//      or
		//       #333221        (rrggbb color)
		//      or ( new format: tk-color plus alpha  - note; no space around ',' ; alpha in hex-format)
		//       red,fe         (symbolic name plus alpha)
		//       #ff00aa,fe     (rrggbb plus alpha)
		%typemap(in) (Color) {				     
            const char *colorStr = Tcl_GetString($input);  
 			
			XColor *colorPtr;
			int alpha = 255;
			const char *sep = strchr(colorStr,',');

			if ( sep ) {
					if( strlen(sep+1) != 2 ) {
		     	   		SWIG_Tcl_SetErrorMsg(interp,"Color specification:", "alpha must be 2 dgits in hex-format");
		           		return TCL_ERROR;
					}
					char *dupStr;
					int len = (sep-colorStr);
					dupStr = (char *)malloc( len+1 );
					strncpy( dupStr, colorStr, len ); dupStr[len] = '\0' ;
					colorPtr = Tk_GetColor(interp,Tk_MainWindow(interp), dupStr);
					free(dupStr);

					char *endStr;
					alpha = (int)strtol(sep+1,&endStr,16);
					if( *endStr != '\0' ) {
		     	   		SWIG_Tcl_SetErrorMsg(interp,"Color specification:", "alpha must be 2 digists in hex-format");
		           		return TCL_ERROR;
		           	}
			} else {
					// normal tk-color
					colorPtr = Tk_GetColor(interp,Tk_MainWindow(interp), colorStr);
			}
			if( !colorPtr ) {
	     	   SWIG_Tcl_SetErrorMsg(interp,"Color specification:", "unknown color name");
	           return TCL_ERROR;
			}
			Tk_FreeColor(colorPtr);

			Color c = Color(colorPtr->red, colorPtr->green, colorPtr->blue, alpha);
		    $1 = c;
		}

	 // WARNING: extracted from the SWIG manual
	 // ... If you are writing your own typemaps for types used in methods with default arguments, you may also need to write a typecheck typemap
	 // If you don't use it there's a conflict between overloaded methods, like the following ...  
	 //		AggKit.i(300): warning 472: Overloaded method AggKit::fillColor() with no explicit typecheck typemap for arg 1 of type 'Color'
	 // APPLY this typecheck for all enums used in methods with default args !!
	%typemap(typecheck) Color = char *;

	%typemap(out) Color {
	   // $1 is a Color
		char formattedStr[10+1];  // 10 is length of #rrggbb,aa
		sprintf( formattedStr, "#%02X%02X%02X,%02X", $1.r, $1.g, $1.b, $1.a );				
		Tcl_Obj *objPtr = Tcl_NewStringObj(formattedStr, -1);
		Tcl_SetObjResult(interp, objPtr);
	}

	// This tells SWIG to treat in-param "RectD" as a special case.
	// i.e.  We have C methods with "RectD" parameter is expected,
	//   and in Tcl we should pass a list of 4 doubles 
	%typemap(in) (RectD) {
	     Tcl_Obj **listobjv;
	     int       nitems;
	     int       i;
	     if (Tcl_ListObjGetElements(interp, $input, &nitems, &listobjv) == TCL_ERROR) {
	        return TCL_ERROR;
	     }
	     if (nitems != 4) {
	     	SWIG_Tcl_SetErrorMsg(interp,"Rectangle specification:", "expected list of 4 floating-point numbers");
	        return TCL_ERROR;
	     }
	     
	     double c[4];
	     double *pc = &c[0];
	     for (i = 0; i < nitems; i++) {
	        if ( Tcl_GetDoubleFromObj(interp,listobjv[i],pc+i) != TCL_OK ) {
	     	   SWIG_Tcl_SetErrorMsg(interp,"Rectangle specification:", "Parameters must be floating-point numbers");
	           return TCL_ERROR;
			}
	     }
	     $1 = RectD(c[0],c[1],c[2],c[3]);
	}

	%typemap(out) RectD {
	   // $1 is a RectD 
		Tcl_Obj *listPtr = Tcl_NewListObj(0, NULL);
	
		Tcl_ListObjAppendElement(NULL, listPtr, Tcl_NewDoubleObj($1.x1));
		Tcl_ListObjAppendElement(NULL, listPtr, Tcl_NewDoubleObj($1.y1));
		Tcl_ListObjAppendElement(NULL, listPtr, Tcl_NewDoubleObj($1.x2));
		Tcl_ListObjAppendElement(NULL, listPtr, Tcl_NewDoubleObj($1.y2));
		Tcl_SetObjResult(interp, listPtr);
	}

	// This tells SWIG to treat in-param "Transformations" as a special case.
	// i.e.  We have C methods with "Transformations" parameter is expected,
	//   and in Tcl we should pass a list of 6 doubles 
	%typemap(in) (Transformations) {
	     Tcl_Obj **listobjv;
	     int       nitems;
	     int       i;
	     if (Tcl_ListObjGetElements(interp, $input, &nitems, &listobjv) == TCL_ERROR) {
	        return TCL_ERROR;
	     }
	     if (nitems != 6) {
	     	SWIG_Tcl_SetErrorMsg(interp,"Matrix Transformation:", "expected list of 6 floating-point numbers");
	        return TCL_ERROR;
	     }
	     
		 Transformations m;
	     double *pc = (double *)&m;
	     for (i = 0; i < nitems; i++) {
	        if ( Tcl_GetDoubleFromObj(interp,listobjv[i],pc+i) != TCL_OK ) {
	     	   SWIG_Tcl_SetErrorMsg(interp,"Matrix Transformation:", "Parameters must be floating-point numbers");
	           return TCL_ERROR;
			}
	     }
	     $1 = m;
	}

	%typemap(out) Transformations {
	   // $1 is a Transformations 
		int i;
		Tcl_Obj *listPtr = Tcl_NewListObj(0, NULL);

		double *p = (double *)&$1;
		for( i=0; i< 6 ; ++i) {
			Tcl_ListObjAppendElement(NULL, listPtr, Tcl_NewDoubleObj(*p++));
		}
		Tcl_SetObjResult(interp, listPtr);
	}

// -----------------

	// This function and the next %typemap allow to pass enum values 
	//  just by using their names (without a leading $)
	// derived from tcl3dConstHash.i
	%{
		static int My_GetConst (Tcl_Interp *interp, Tcl_Obj *obj, long *valPtr) {
			int     len;
			Tcl_Obj *foundObj = NULL;
			char *str = Tcl_GetStringFromObj (obj, &len);
			foundObj = SWIG_GetConstant (str);
			if (!foundObj) {
				return Tcl_GetLongFromObj (interp, obj, valPtr);
			}
			return Tcl_GetLongFromObj (interp, foundObj, valPtr);
		}
	%}


// TO DO: check if all enum are listed ...
	// Change input handling to look for names for the specified enums
	%typemap(in) 
	      enum Linejoin, enum LineCap
		, enum ViewportOption, enum Direction
		, enum DrawPathFlag
		, enum TextAlignment, enum FontCacheType
		, enum ImageFilter, enum ImageResample, enum BlendMode		
	{
	    long temp; 
	    if (My_GetConst (interp, $input, &temp) == TCL_ERROR) {
	        return TCL_ERROR;
	    } else {
	        $1 = ($type) temp;
	    }
	}

// -------------
	enum LineCap {
		CapButt   = agg::butt_cap,
		CapSquare = agg::square_cap,
		CapRound  = agg::round_cap
	};
    enum LineJoin {
        JoinMiter = agg::miter_join,
        JoinRound = agg::round_join,
        JoinBevel = agg::bevel_join
    };	
	enum DrawPathFlag {
		FillOnly,
		StrokeOnly,
		FillAndStroke,
		FillWithLineColor
	};
	enum Direction {
		CW, CCW
	};
    enum FontCacheType {
        RasterFontCache,
        VectorFontCache
    };
    enum TextAlignment {
        AlignLeft,
        AlignRight,
        AlignCenter,
        AlignTop = AlignRight,
        AlignBottom = AlignLeft
    };
	enum ViewportOption	{
		Anisotropic,
		XMinYMin,
		XMidYMin,
		XMaxYMin,
		XMinYMid,
		XMidYMid,
		XMaxYMid,
		XMinYMax,
		XMidYMax,
		XMaxYMax
	};
    enum ImageFilter {
        NoFilter,
        Bilinear,
        Hanning,
        Hermite,
        Quadric,
        Bicubic,
        Catrom,
        Spline16,
        Spline36,
        Blackman144
    };
    enum ImageResample {
        NoResample,
        ResampleAlways,
        ResampleOnZoomOut
    };
    enum BlendMode {
        BlendAlpha      = agg::end_of_comp_op_e,
        BlendClear      = agg::comp_op_clear,
        BlendSrc        = agg::comp_op_src,
        BlendDst        = agg::comp_op_dst,
        BlendSrcOver    = agg::comp_op_src_over,
        BlendDstOver    = agg::comp_op_dst_over,
        BlendSrcIn      = agg::comp_op_src_in,
        BlendDstIn      = agg::comp_op_dst_in,
        BlendSrcOut     = agg::comp_op_src_out,
        BlendDstOut     = agg::comp_op_dst_out,
        BlendSrcAtop    = agg::comp_op_src_atop,
        BlendDstAtop    = agg::comp_op_dst_atop,
        BlendXor        = agg::comp_op_xor,
        BlendAdd        = agg::comp_op_plus,
        BlendMultiply   = agg::comp_op_multiply,
        BlendScreen     = agg::comp_op_screen,
        BlendOverlay    = agg::comp_op_overlay,
        BlendDarken     = agg::comp_op_darken,
        BlendLighten    = agg::comp_op_lighten,
        BlendColorDodge = agg::comp_op_color_dodge,
        BlendColorBurn  = agg::comp_op_color_burn,
        BlendHardLight  = agg::comp_op_hard_light,
        BlendSoftLight  = agg::comp_op_soft_light,
        BlendDifference = agg::comp_op_difference,
        BlendExclusion  = agg::comp_op_exclusion,
    };

%include <typemaps.i>


%rename(Pixmap) AggPixmap;
  // usage:  Agg::Pixmap ....
class AggPixmap {
  public:
  	AggPixmap();
    AggPixmap(unsigned char* buf, unsigned width, unsigned height, int stride);
    void attach(unsigned char* buf, unsigned width, unsigned height, int stride);    
    int width();
    int height();
    int stride();
    unsigned char* buf();
    void premultiply();
    void demultiply();
	void swapRxB();
	 // WARNING: subImage should be used only in temporary contexts;
	 //  Since subImage shares the pixelbuffer with he 'original' Pixmap,
	 //  and the original Pixmap may reallocate its buffer, then 
	 //  subImage's shared buffer may become invalid !!! 
	AggPixmap subImage(unsigned int x0, unsigned int y0, unsigned int dx, unsigned int dy);
	void copy( AggPixmap &src );
	void clear ();
};




%include <attribute.i>
	// attributes are converted in tcl widgets options
 	//  Note that attribute xxx will be named -xxx

%attribute(AggTcl, RectD, clipbox, clipBox, clipBox);

%attribute(AggTcl, enum BlendMode, blendmode, blendMode, blendMode);
%attribute(AggTcl, enum BlendMode, imageblendmode, imageBlendMode, imageBlendMode);

%attribute(AggTcl, Color, fillcolor, fillColor, fillColor);
%attribute(AggTcl, Color, linecolor, lineColor, lineColor);
%attribute(AggTcl, Color, imageblendcolor, imageBlendColor, imageBlendColor);

%attribute(AggTcl, double, masteralpha, masterAlpha, masterAlpha);
%attribute(AggTcl, double, antialiasgamma, antiAliasGamma, antiAliasGamma);

%attribute(AggTcl, double, linewidth, lineWidth, lineWidth);
%attribute(AggTcl, enum LineCap, linecap, lineCap, lineCap);
%attribute(AggTcl, enum LineJoin, linejoin, lineJoin, lineJoin);
 // non mi piace il nome
%attribute(AggTcl, bool, fillevenodd, fillEvenOdd, fillEvenOdd);

%attribute(AggTcl, bool, texthints, textHints, textHints);
%attribute(AggTcl, bool, textflip, flipText, flipText);
%attribute(AggTcl, enum TextAlignment, textalignx, textAlignX, textAlignX);
%attribute(AggTcl, enum TextAlignment, textaligny, textAlignY, textAlignY);

%attribute(AggTcl, enum ImageFilter, imagefilter, imageFilter, imageFilter);
%attribute(AggTcl, enum ImageResample, imageresample, imageResample, imageResample);

%attribute(AggTcl, Transformations, transformations, transformations, transformations);

 // swig - trick : 
 //  define here the specialized class AggTcl *PLUS* all the inherithed methods (from AggKit))    
 //  NOTE: the 'real (base) class' AggKit is pure-C (no relation with OS-graphics, nor with tcl interpreter).
 //    AggTcl is just used as a swig wrapper, extending the base class, and
 //    adding support for the tcl environment.

 
%rename(Agg) AggTcl;

class AggTcl {
  public:
  	void updated();
	void setUpdateScript(Tcl_Interp * interp, const char *script);
	  	 
	// *explicit* all the inherited methods (from AggKit class) you want to wrap ..
	  
    // Setup
    //-----------------------
	void attach(unsigned char *buf, unsigned int w, unsigned int h, int stride);
    void  attach(AggPixmap& img);
    int width();
    int height();
	void clear(Color c);
    void clearClipBox(Color c);

	 // WARNING: subImage should be used only in temporary contexts;
	 //  Since subImage shares the pixelbuffer with he 'original' Pixmap,
	 //  and the original Pixmap may reallocate its buffer, then 
	 //  subImage's shared buffer may become invalid !!! 
	AggPixmap subImage();
	AggPixmap subImage(unsigned int x0, unsigned int y0, unsigned int dx, unsigned int dy);

    
    // Conversions
    //-----------------------
		// make double& INOUT
		%apply double& INOUT {double &x}
		%apply double& INOUT {double &y}
    
    void   worldToScreen(double &x, double &y) const;
    void   screenToWorld(double &x, double &y) const;

    bool   inBox(double worldX, double worldY) const;    
    
    // Transformations
    //-----------------------
	 // NOTE: you can get/set transformations through configuration option -transformations
    void resetTransformations();
    void rotate(double angle);
    void scale(double sx, double sy);
    void skew(double sx, double sy);
    void translate(double x, double y);

	 // WARNING: extracted from the SWIG manual
	 // ... If you are writing your own typemaps for types used in methods with default arguments, you may also need to write a typecheck typemap
	 // APPLY this typecheck for all enums used in methods with default args !!
	%typemap(typecheck) enum ViewportOption = char *;
    void viewport(double worldX1,  double worldY1,  double worldX2,  double worldY2,
                  double screenX1, double screenY1, double screenX2, double screenY2,
                  enum ViewportOption opt=XminYmin);

    void fillLinearGradient(double x1, double y1, double x2, double y2, Color c1, Color c2, double profile=1.0);
    void lineLinearGradient(double x1, double y1, double x2, double y2, Color c1, Color c2, double profile=1.0);

    void fillRadialGradient(double x, double y, double r, Color c1, Color c2, double profile=1.0);
    void lineRadialGradient(double x, double y, double r, Color c1, Color c2, double profile=1.0);

    void fillRadialGradient(double x, double y, double r, Color c1, Color c2, Color c3);
    void lineRadialGradient(double x, double y, double r, Color c1, Color c2, Color c3);

    void fillRadialGradient(double x, double y, double r);
    void lineRadialGradient(double x, double y, double r);


    // Basic Shapes
    //-----------------------
    void line(double x1, double y1, double x2, double y2);
    void triangle(double x1, double y1, double x2, double y2, double x3, double y3);
    void rectangle(double x1, double y1, double x2, double y2);
    void roundedRect(double x1, double y1, double x2, double y2, double r);
    void roundedRect(double x1, double y1, double x2, double y2, double rx, double ry);
    void roundedRect(double x1, double y1, double x2, double y2,
                     double rxBottom, double ryBottom,
                     double rxTop,    double ryTop);   
    void ellipse(double cx, double cy, double rx, double ry);
    void arc(double cx, double cy, double rx, double ry, double start, double sweep);
    void star(double cx, double cy, double r1, double r2, double startAngle, int numRays);
    void curve(double x1, double y1, double x2, double y2, double x3, double y3);
    void curve(double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4);

		// This tells SWIG to treat in-param "double *xy" as a special case
		%typemap(in) (double *xy, int numPoints) {
		     Tcl_Obj **listobjv;
		     double    *vv;
		     int       nitems;
		     int       i;
		     if (Tcl_ListObjGetElements(interp, $input, &nitems, &listobjv) == TCL_ERROR) {
		        return TCL_ERROR;
		     }
		     
		     vv = (double *) malloc(nitems*sizeof(double));
		     for (i = 0; i < nitems; i++) {
		        if ( Tcl_GetDoubleFromObj(interp,listobjv[i],vv+i) != TCL_OK ) {
		        	free(vv);
		           return TCL_ERROR;
		           //  ?? nothing is loaded, print an error ??? 
				}
		     }
		     $1 = vv;
		     $2= nitems;
		}
		// This gives SWIG some cleanup code that will get called after the function call
		%typemap(freearg) (double *xy, int numPoints) {
		     if ($1) {
		        free($1);
		     }
		}

    void polygon(double* xy, int numPoints);
    	// NOTE: Tcl usage:   $obj polygon {0.0 0.0 1.1 1.0 2.0 4.0}
    void polyline(double* xy, int numPoints);
    	// NOTE: Tcl usage:   $obj polyline {0.0 0.0 1.1 1.0 2.0 4.0}


    // Text
    //-----------------------
	 // WARNING: extracted from the SWIG manual
	 // ... If you are writing your own typemaps for types used in methods with default arguments, you may also need to write a typecheck typemap
	 // If you don't use it there's a conflict between overloaded methods, like the following ...  
	 //		AggKit.i(300): warning 472: Overloaded method AggKit::fillColor() with no explicit typecheck typemap for arg 1 of type 'Color'
	%typemap(typecheck) enum FontCacheType = char *;
    void   font(const char* fileName, double height,
                bool bold = false,
                bool italic = false,
                enum FontCacheType ch = RasterFontCache,
                double angle = 0.0);
    double fontHeight() const;
    double textWidth(const char* str);
    void   text(double x, double y, const char* str, bool roundOff=false, double dx=0.0, double dy=0.0);

    // Path commands
    //-----------------------
    void resetPath();

    void moveTo(double x, double y);
    void moveRel(double dx, double dy);

    void lineTo(double x, double y);
    void lineRel(double dx, double dy);

    void horLineTo(double x);
    void horLineRel(double dx);

    void verLineTo(double y);
    void verLineRel(double dy);

    void arcTo(double rx, double ry,
               double angle,
               bool largeArcFlag,
               bool sweepFlag,
               double x, double y);

    void arcRel(double rx, double ry,
                double angle,
                bool largeArcFlag,
                bool sweepFlag,
                double dx, double dy);

    void quadricCurveTo(double xCtrl, double yCtrl, double xTo,   double yTo);
    void quadricCurveRel(double dxCtrl, double dyCtrl, double dxTo,   double dyTo);
    void quadricCurveTo(double xTo, double yTo);
    void quadricCurveRel(double dxTo, double dyTo);

    void cubicCurveTo(double xCtrl1, double yCtrl1,
                      double xCtrl2, double yCtrl2,
                      double xTo,    double yTo);

    void cubicCurveRel(double dxCtrl1, double dyCtrl1,
                       double dxCtrl2, double dyCtrl2,
                       double dxTo,    double dyTo);

    void cubicCurveTo(double xCtrl2, double yCtrl2, double xTo,    double yTo);

    void cubicCurveRel(double xCtrl2, double yCtrl2, double xTo,    double yTo);

    void addEllipse(double cx, double cy, double rx, double ry, Direction dir);
    void closePolygon();

	 // WARNING: extracted from the SWIG manual
	 // ... If you are writing your own typemaps for types used in methods with default arguments, you may also need to write a typecheck typemap
	 // If you don't use it there's a conflict between overloaded methods, like the following ...  
	 //		AggKit.i(300): warning 472: Overloaded method AggKit::fillColor() with no explicit typecheck typemap for arg 1 of type 'Color'
	 // APPLY this typecheck for all enums used in methods with default args !!
	%typemap(typecheck) enum DrawPathFlag = char *;    
	void drawPath(enum DrawPathFlag flag = FillAndStroke);

        // Image Transformations
    //-----------------------
    void paintImage(const AggPixmap& img,
                           int imgX1,    int imgY1,    int imgX2,    int imgY2,
                        double dstX1, double dstY1, double dstX2, double dstY2);

    void paintImage(const AggPixmap& img,
                        double dstX1, double dstY1, double dstX2, double dstY2);

    void paintImageInPath(const AggPixmap& img,
                           int imgX1,    int imgY1,    int imgX2,    int imgY2,
                        double dstX1, double dstY1, double dstX2, double dstY2);

    void paintImageInPath(const AggPixmap& img,
                        double dstX1, double dstY1, double dstX2, double dstY2);

    // Image Blending (no transformations available)
    void blendImage(AggPixmap& img,
                    int imgX1, int imgY1, int imgX2, int imgY2,
                    double dstX, double dstY, unsigned alpha);
    void blendImage(AggPixmap& img, double dstX, double dstY, unsigned alpha);
};



