%module TkPixUtil

/* 
   ==========================================================================
 This module provides a tcl interface to some C functions useful for 
  managing a minimal GDI (Windows) graphic context,
  alloc/dealloc pixmaps (matrix of pixels)
  transforming pixmap in tk-photo
  ---
  Aug 2015 - A.Buratti 

  GetDC _hWnd_
    Returns an opaque handler of the DeviceContext (HDC).
	_hWnd_ is a Window-specific identifier for a window, 
	as returned by [winfo id _window_]

  ReleaseDC _hWnd_ _HDC_
    Release the DeviceContext _HDC_ ( see GetDC command above).
	_hWnd_ should be the same Window-identifier used for getting _HDC_.


  malloc_buffer _n_
    Allocates a buffer of _n_ bytes. Returns an opaque handle pointing to the buffer.
    If there is not enough menory, returns "NULL".
  free_buffer _handle_
    Frees the memory allocated with [malloc_buffer..]
	
  malloc_BITMAPINFOHEADER
  	Allocates a buffer for a BITMAPINFOHEADER. 
	Returns an opaque handle pointing to the buffer.
    If there is not enough menory, returns "NULL".
  free_BITMAPINFOHEADER _BITMAPINFO_handle_
    Frees the memory allocated with [malloc_BITMAPINFOHEADER..]

  setBITMAPINFOHEADER _BITMAPINFOHEADER_handle_  _width_ _height_  _bpp_
    sets the BITMAPINFOHEADER structure.
    This structure is required for blitting (see next), i.e. for displaying a
	pixel-matrix in a window.
    This structure describes how the bytes of a raw-buffer (see malloc_buffer)
    are organized to form a pixel-matrix. 
	You only need to pass _width_, _height_, _bpp_ (bitsPerPixels)
    * _bpp_ must be 32 * ; other internal fields are 'hardcoded' by this
    procedure. 

  blitAll _HDC_ _buffer_handle_ _BITMAPINFO_handle_
    Paints the pixel-matrix located by _buffer_handle_ in the window located by _HDC_.
	_BITMAPINFO_handle_ must describe how the raw-memory of _buffer_handle_ is
	organized to form a pixel-matrix.
	If the size of the pixel-matrix exceeds the window-area, it is clipped.

  blitRegion _HDC_ _buffer_handle_ _BITMAPINFO_handle_ _x_ _y_ _dx_ _dy_
    Similar to blitAll, but only the region (x,y,dx,dy) is repainted.
	 

  TkPhotoInfo _photoName_ 
   Returns info about a tk-photo or raises an error if _photoName_
    does not exist.
   Info is returned as a tcl dictionary with the following entries:
	 width  - photo's width
	 height - photo's height
	 pixelSize - pixel size (in bytes) [it must be 4]
	 pitch - the address difference between two vertically adjacent pixels.
	 pixelPtr - an opaque handle pointing to the 1st pixel (or NULL)
	 pixelAddress - the memory address (as integer) of the first pixel

 
   Tk_Pixmap2Photo _pPtr_ _pDx_ _pDy_ _pPitch_
                   _x0_  _y0_ _dx_ _dy_
                   _photoName_
    This proc copies a block of pixels in a tk-photo.
    Basically, it is a wrapping of Tk_PhotoPutBlock ... 
     pPtr, pDx, pDy, pStride define the input block of pixels.
    We assume that each pixel is 4 bytes, formatted as ABGR (windows).
    pPtr must be an opaque handle (as those returned by malloc_buffer) 
    pointing to the first pixel, that is, the top-left pixel in the block.
    pDx and pDY specify the dimensions of the block of pixels.
    pPitch specifies the address difference between two vertically adjacent pixels. 
    x0 y0 dx dy define a rectangular sub-region of the destination image (_photName_).

    Tk_Pixmap2Photo raises an error if _photoName_ does not exist.
    
	For further details se the "Tk_PhotoPutBlock" manual page ..

	NOTES:
	 See AggTk.tcl for a complete usage of these functions.
	
	*WINDOWS-GDI and other graphic-platforms*
	Note that the only parts dealing with Windows(GDI) are three very simple procs:
	* GetDC, ReleaseDC, blitAll/blitRegion (plus the auxiliary setBITMAPINFOHEADER).
	It should not be hard to write similar functions for other graphic platforms ..	 	
	   
   ==========================================================================
*/


%{
#include <Windows.h> 
#include <tk.h>

  // DANGER ! *never* build error-messages with more than 250 chars !!!
static char ExceptionBuffer[250];

%}


%typemap(in) HWND {
	if (Tcl_GetLongFromObj(interp,$input,(long *)&$1) == TCL_ERROR) return TCL_ERROR;
}
HDC     GetDC( HWND hWnd );
int     ReleaseDC( HWND hWnd, HDC hDC );



%include <cmalloc.i>
%malloc(unsigned char,buffer);  //  -> generates: unsigned char * malloc_buffer(int size)
%free(unsigned char,buffer);    //  -> generates: void free_buffer(unsigned char*buf)


%malloc(BITMAPINFOHEADER);  //  -> generates: unsigned char * malloc_BITMAPINFOHEADER(int size=sizeof(BITMAPINFOHEADER))
%free(BITMAPINFOHEADER);    //  -> generates: void free_BITMAPINFOHEADER(BITMAPINFOHEADER *buf)
%inline {
	void setBITMAPINFOHEADER ( BITMAPINFOHEADER *p, int width, int height, int bpp ) {
		p->biSize = sizeof(BITMAPINFOHEADER);
		p->biWidth = width;
		p->biHeight = height;
		p->biPlanes = 1;
		p->biBitCount = bpp;
		p->biCompression = BI_RGB;
		p->biSizeImage = 0; // may be zero for BI_RGB
		p->biXPelsPerMeter = 0;
		p->biYPelsPerMeter = 0;
		p->biClrUsed = 0;
		p->biClrImportant = 0;
	}

	void blitAll( HDC hDC, unsigned char *pixelPtr, BITMAPINFOHEADER *info ) {
		int h = abs(info->biHeight);
		SetDIBitsToDevice( hDC, 0, 0, info->biWidth, h,  0, 0, 0, h, pixelPtr, (BITMAPINFO *)info, DIB_RGB_COLORS );
	}

	void blitRegion ( HDC hDC, unsigned char *pixelPtr, BITMAPINFOHEADER *info, int x, int y , int dx, int dy ) {
		int h = abs(info->biHeight);
		SetDIBitsToDevice( hDC, x, y, dx, dy,  x, h-(dy+y) , 0, h, pixelPtr, (BITMAPINFO *)info, DIB_RGB_COLORS );
	}
}


%include <tclinterp.i>

%typemap(out) Tk_PhotoImageBlock {
   // transform a Tk_PhotoImageBlock in a Tcl-dictionary 
	
	Tcl_Obj *dictObj = Tcl_NewDictObj();

	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("width",-1),
		Tcl_NewIntObj($1.width) );
	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("height",-1),
		Tcl_NewIntObj($1.height) );
	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("pixelSize",-1),
		Tcl_NewIntObj($1.pixelSize) );
	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("pitch",-1),
		Tcl_NewIntObj($1.pitch) );
	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("pixelPtr",-1),
		SWIG_NewInstanceObj( SWIG_as_voidptr($1.pixelPtr), SWIGTYPE_p_unsigned_char,0) );
	Tcl_DictObjPut(interp, dictObj,
		Tcl_NewStringObj("pixelAddress",-1),
		Tcl_NewLongObj((long)$1.pixelPtr) );
	Tcl_SetObjResult(interp, dictObj);
}

%inline {

	typedef char *TkPixException;
	
	 // returns info deep-level info about a tk-photo (imageName).
	 // If imageName does not exist, it will raise an error.      
   	Tk_PhotoImageBlock TkPhotoInfo( Tcl_Interp *interp, char *photoName ) throw(TkPixException) {
		Tk_PhotoHandle photoHandle;
		Tk_PhotoImageBlock imageBlock;

		photoHandle = Tk_FindPhoto(interp, photoName);
		if (photoHandle == 0) {
		     // DANGER ! you must always limit the lenght of formatted params !!
			sprintf(ExceptionBuffer, "image \"%.20s\" does not exist.", photoName);
			throw(ExceptionBuffer);
		}
		Tk_PhotoGetImage(photoHandle, &imageBlock);
		return imageBlock;	
	}

 
	 // This function copies a block of pixels in a tk-photo.
	 // Basically, it is a wrapping of Tk_PhotoPutBlock ... 
	 // pPtr, pDx, pDy, pStride define the input block of pixels.
	 //   We assume that each pixel is 4 bytes, formatted as ABGR (windows)
	 //   pPtr points to the first pixel, that is, the top-left pixel in the block.
	 //   pDx and pDY specify the dimensions of the block of pixels.
	 //   pPitch specifies the address difference between two vertically adjacent pixels. 
	 // x0 y0 dx dy defines a rectangular sub-region of the destination image (imageName).
	 // For further details se the "Tk_PhotoPutBlock" manual page ..

	int Tk_Pixmap2Photo(Tcl_Interp *interp, 
				unsigned char *pPtr, unsigned int pDx, unsigned int pDy, int pPitch,
				unsigned int x0, unsigned int y0, unsigned int dx, unsigned int dy,
				char *imageName) throw(TkPixException) {
		Tk_PhotoHandle photoHandle;
		photoHandle = Tk_FindPhoto(interp, imageName);
		if (photoHandle == 0) {
		     // DANGER ! you must always limit the lenght of formatted params !!
			sprintf(ExceptionBuffer, "image \"%.20s\" does not exist.", imageName);
			throw(ExceptionBuffer);
		}		
		Tk_PhotoImageBlock tkBlockPtr;
			tkBlockPtr.pixelPtr = pPtr ;
			tkBlockPtr.width = pDx;
			tkBlockPtr.height = pDy;
			tkBlockPtr.pitch = pPitch;
			tkBlockPtr.pixelSize = 4;
			 // WARNING: swap red and blue components ..
			tkBlockPtr.offset[0] = 2 ;
			tkBlockPtr.offset[1] = 1 ;
			tkBlockPtr.offset[2] = 0 ;
			tkBlockPtr.offset[3] = 3 ;

		return Tk_PhotoPutBlock(interp, photoHandle, &tkBlockPtr, x0, y0, dx, dy, TK_PHOTO_COMPOSITE_SET);
	}
}

